#!/usr/bin/python
# encoding: utf-8
#
# Copyright (c) 2019 Dean Jackson <deanishe@deanishe.net>
#
# MIT Licence. See http://opensource.org/licenses/MIT
#
# Created on 2019-06-08
#

"""Filter Firefox bookmarks."""

from __future__ import print_function, absolute_import

from collections import namedtuple
from ConfigParser import SafeConfigParser
from contextlib import contextmanager
from HTMLParser import HTMLParser
import json
import os
import sys
from time import time
from urlparse import urlparse


# CACHEDIR = os.getenv('alfred_workflow_cache')
# Firefox's data directory
FFDIR = os.path.expanduser('~/Library/Application Support/Firefox')
# Name of profile to load. If unset, default profile is loaded.
PROFILE = os.getenv('profile_name')


def log(s, *args):
    """Log to STDERR."""
    if args:
        s = s % args
    if isinstance(s, unicode):
        s = s.encode('utf-8')
    print(s, file=sys.stderr)


@contextmanager
def timed(title):
    """Log execution time."""
    start = time()
    yield
    log(u'%0.3fs \U000029D7 %s', time() - start, title)


# Firefox user profile
Profile = namedtuple('Profile', 'name path default')
# Firefox bookmark
Bookmark = namedtuple('Bookmark', 'title url')


class Parser(HTMLParser):
    """HTML parser for ``bookmarks.html`` file."""

    @classmethod
    def read(cls, bookmarks_html):
        """Extract links in HTML.

        Args:
            bookmarks_html (HTML or filepath): HTML or HTML file
                to extract bookmarks from.

        Returns:
            list: Bookmarks extracted from HTML.

        """
        if os.path.exists(bookmarks_html):
            with open(bookmarks_html, 'rb') as fp:
                bookmarks_html = fp.read()

        p = cls()
        p.feed(bookmarks_html)
        return sorted(p.bookmarks)

    def __init__(self):
        """Create new parser."""
        self.reset()
        self.bookmarks = set()
        self.title = None
        self.url = None

    def handle_starttag(self, tag, attr):
        """Extract new ``<a>`` tag."""
        if tag == 'a':
            for name, value in attr:
                if name == 'href':
                    self.url = value
                    break

    def handle_endtag(self, tag):
        """Save extracted data."""
        if tag == 'a':
            if self._validate():
                self.bookmarks.add(Bookmark(self.title, self.url))
            self.title = None
            self.url = None

    def handle_data(self, data):
        """Save tag content."""
        self.title = data.strip()

    def _validate(self):
        """Return `True` if parser has found a valid bookmark."""
        if not self.title or not self.url:
            return False
        if self.url.lower().startswith('javascript:'):  # ignore bookmarklets
            return False
        return True


def urldomain(url):
    """Return domain extracted from URL."""
    p = urlparse(url)
    return p.hostname


def _profile_sort_key(p):
    """Return sort key for `Profile`."""
    n = 0 if p.default else 1
    return (n, p.name)


def read_profiles():
    """Load Firefox profiles."""
    profiles = []
    path = os.path.join(FFDIR, 'profiles.ini')
    parser = SafeConfigParser()
    parser.read(path)
    for sect in parser.sections():
        if not sect.startswith('Profile'):
            continue

        default = False
        name = parser.get(sect, 'Name')
        p = parser.get(sect, 'Path')
        if parser.has_option(sect, 'IsRelative') \
                and parser.getboolean(sect, 'IsRelative'):
            p = os.path.join(FFDIR, p)
        if parser.has_option(sect, 'Default'):
            default = parser.getboolean(sect, 'Default')

        profiles.append(Profile(name, p, default))

    profiles.sort(key=_profile_sort_key)

    return profiles


def get_profile():
    """Load user-specified profile or default."""
    profiles = read_profiles()
    if PROFILE:  # load user-specified profile
        for p in profiles:
            if p.name == PROFILE:
                return p

    # return default profile
    for p in profiles:
        if p.default:
            return p


def send_items(*items):
    """Send Alfred JSON."""
    json.dump(dict(items=items), sys.stdout, indent=2, sort_keys=True)


def main():
    """Run Script Filter."""
    log('.')
    with timed('load profile'):
        profile = get_profile()
    path = os.path.join(profile.path, 'bookmarks.html')
    log('profile=%r', profile)
    with timed('load bookmarks'):
        bookmarks = Parser.read(path)

    items = []
    for bm in bookmarks:
        match = bm.title
        domain = urldomain(bm.url)
        if domain:
            match += ' ' + domain
        items.append(dict(
            title=bm.title,
            subtitle=bm.url,
            match=match,
            arg=bm.url,
            uid=bm.url,
            valid=True,
        ))

    log('%d bookmark(s) in profile "%s"', len(items), profile.name)

    send_items(*items)
    return 0


if __name__ == "__main__":
    try:
        with timed('done'):
            code = main()
        sys.exit(code)
    except Exception as err:
        import traceback
        log('[ERROR] %s', err)
        log('-------------------- TRACEBACK --------------------')
        traceback.print_exc()
        log('------------------ END TRACEBACK ------------------')
        sys.exit(1)
